#include "cui.h"
#include "focus.h"
#include "rect.h"
#include "esc.h"
#include "key.h"
#include "handler.h"
#include "panel.h"
#include "label.h"
#include "button.h"
#include "timer.h"
#include "list.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <termios.h>

typedef struct cui_main_stack *cui_main_stack;

struct cui_main_stack{
	cui focus;
	int running;
	cui_main_stack prev;
};

static cui_main_stack main_stack = NULL;
cui_rect cui_clip = NULL;

cui
cui_focus_get(void)
{
	return main_stack ? main_stack->focus : NULL;
}

void
cui_focus_set(cui obj)
{
	if(main_stack) main_stack->focus = obj;
}

void
cui_focus_chk(void)
{
	cui f = cui_focus_get();
	if(f && (!cui_is_visible(f) || !cui_is_visible_rect(f))) cui_focus_set(NULL);
}

int
cui_running_get(void)
{
	return main_stack ? main_stack->running : FALSE;
}

void
cui_running_set(int v)
{
	if(main_stack) main_stack->running = v;
}

cui
cui_alloc(int size)
{
	cui obj;

	if((obj = malloc(size)) == NULL) ERR("No Mem");
	return obj;
}

void
cui_base_init(cui obj, cui parent, int x, int y, int w, int h)
{
	obj->parent = parent;
	obj->children = NULL;
	obj->next = NULL;
	obj->x = x;
	obj->y = y;
	obj->w = w;
	obj->h = h;

	if(parent) cui_base_child_add(parent, obj);
	obj->handler_list = NULL;
	obj->flags = 0;
}

void
cui_base_child_add(cui obj, cui child)
{
	child->next = obj->children;
	obj->children = child;
}

cui
cui_base_new(cui parent, int x, int y, int w, int h)
{
	cui obj = cui_alloc(sizeof(struct cui_base));
	cui_base_init(obj, parent, x, y, w, h);
	return obj;
}

void
cui_draw_str(cui obj, int x, int y, char *s, int attr)
{
	struct cui_rect t1, t2;
	cui_rect r1 = &t1, r2 = &t2;
	int n = strlen(s);
	char *s2 = NULL;

	if(!cui_is_visible(obj)) return;
	if(!cui_visible_rect(obj, r1)) return;
	cui_rect_init(r2, x, y, n, 1);
	if(!cui_rect_and(r1, r2, r2)) return; /* r1 and r2 --> r2 */
	if(cui_clip){
		*r1 = *cui_clip;
		r1->x = cui_lx(obj, r1->x);
		r1->y = cui_ly(obj, r1->y);
		if(!cui_rect_and(r1, r2, r2)) return;
	}

	s += r2->x - x;
	if(r2->x + r2->w < x + n){
		n = r2->w;
		if((s2 = malloc(n + 1)) == NULL) ERR("No Mem");
		memcpy(s2, s, n);
		s2[n] = '\0';
		s = s2;
	}
	x = r2->x;
	cui_esc_loc(cui_gx(obj) + x, cui_gy(obj) + y);
	if(attr & CUI_ATTR_ULINE) cui_esc_attr(CUI_ESC_ULINE);
	if(attr & CUI_ATTR_REVERSE) cui_esc_attr(CUI_ESC_REVERSE);
	printf("%s", s);
	if(attr != CUI_ATTR_NORMAL) cui_esc_attr(CUI_ESC_NORMAL);
	fflush(stdout);
	if(s2) free(s2);
}

void
cui_fill_rect(cui obj, int x, int y, int w, int h, char *s, int attr)
{
	int ix, iy;
	int n = strlen(s);
	char *buf;

	if(w <=0 || h <= 0) return;
	if((buf = malloc(w + 1)) == NULL) ERR("No Mem");
	for(ix=0; ix<w; ix+=n) buf[ix] = s[ix%n];
	buf[w] = '\0';
	for(iy=0; iy<h; iy++) cui_draw_str(obj, x, y+iy, buf, attr);
	free(buf);
}

void
cui_clear_rect(cui obj, int x, int y, int w, int h)
{
	cui_fill_rect(obj, x, y, w, h, " ", CUI_ATTR_NORMAL);
}

void
cui_clear(cui obj)
{
	cui_clear_rect(obj, 0, 0, obj->w, obj->h);
}

int
cui_gx(cui obj)
{
	return obj == NULL ? 0 : cui_gx(obj->parent) + obj->x;
}

int
cui_gy(cui obj)
{
	return obj == NULL ? 0 : cui_gy(obj->parent) + obj->y;
}

int
cui_glen(cui o1, cui o2)
{
	int dx, dy;

	dx = (cui_gx(o1) + o1->w / 2) - (cui_gx(o2) + o2->w / 2);
	dy = (cui_gy(o1) + o1->h / 2) - (cui_gy(o2) + o2->h / 2);
	dy *= 2;
	return dx * dx + dy * dy;
}

int
cui_lx(cui obj, int gx)
{
	return obj == NULL ? gx : cui_lx(obj->parent, gx) - obj->x;
}

int
cui_ly(cui obj, int gy)
{
	return obj == NULL ? gy : cui_ly(obj->parent, gy) - obj->y;
}

int
cui_x2(cui obj)
{
	return obj->x + obj->w;
}

int
cui_y2(cui obj)
{
	return obj->y + obj->h;
}

cui
cui_root(cui obj)
{
	return obj->parent == NULL ? obj : cui_root(obj->parent);
}

void
cui_draw(cui obj)
{
	int val;
	
	if(obj == NULL) return;
	if(!cui_is_visible(obj)) return;
	val = (obj == cui_focus_get()) ? CUI_DRAW_FOCUS : CUI_DRAW_NORMAL;
	cui_handler_call(obj, CUI_EVT_DRAW, val);
	//usleep(20*1000); // for debug
	cui_draw_chain(obj->children);
}

void
cui_draw_chain(cui obj)
{
	if(obj == NULL) return;
	cui_draw_chain(obj->next);
	cui_draw(obj);
}

void
cui_hide(cui obj)
{
	if(obj->flags & CUI_FLG_HIDE) return;
	if(cui_esc_cnt > 0) cui_clear(obj);
	obj->flags |= CUI_FLG_HIDE;
}

void
cui_show(cui obj)
{
	if(!(obj->flags & CUI_FLG_HIDE)) return;
	obj->flags &= ~CUI_FLG_HIDE;
	if(cui_esc_cnt > 0) cui_draw(obj);
}

int
cui_is_visible(cui obj)
{
	if(obj == NULL) return TRUE; /* root */
	return !(obj->flags & CUI_FLG_HIDE) && cui_is_visible(obj->parent);
}

int
cui_visible_rect(cui obj, cui_rect r)
{
	struct cui_rect tmp;
	cui_rect t = &tmp;

	cui_rect_init(r, 0, 0, obj->w, obj->h);
	if(obj->parent == NULL) return !cui_rect_chk_empty(r); /* empty --> FALSE */
	if(!cui_visible_rect(obj->parent, t)) return FALSE;
	t->x -= obj->x;
	t->y -= obj->y;
	return cui_rect_and(r, t, r); /* r and t --> r */
}

int
cui_is_visible_rect(cui obj)
{
	struct cui_rect tmp;
	return cui_visible_rect(obj, &tmp);
}

int
cui_hide_enter(cui obj)
{
	int bak = cui_is_visible(obj);
	if(bak) cui_hide(obj);
	return bak;
}

void
cui_hide_exit(cui obj, int bak)
{
	if(bak) cui_show(obj);
}

void
cui_xywh_set(cui obj, int x, int y, int w, int h)
{
	int visi = cui_is_visible(obj) && cui_esc_cnt > 0;
	int rsz = (obj->w != w || obj->h != h);
	struct cui_rect r1, r2;
	cui_rects rs, ri;

	if(visi){
		if(obj->parent) cui_rect_init(&r1, cui_gx(obj), cui_gy(obj), obj->w, obj->h);
		else cui_clear(obj);
		obj->flags |= CUI_FLG_HIDE;
	}
	obj->x = x;
	obj->y = y;
	obj->w = w;
	obj->h = h;
	if(rsz) cui_handler_call(obj, CUI_EVT_RESIZE, TRUE);
	if(visi){
		obj->flags &= ~CUI_FLG_HIDE;
		if(obj->parent){
			cui_rect_init(&r2, cui_gx(obj), cui_gy(obj), obj->w, obj->h);
			rs = cui_rect_or(&r1, &r2);
			for(ri=rs; ri; ri=ri->next){
				cui_clip = (cui_rect)ri;
				cui_clear(obj->parent);
				cui_draw(obj->parent);
			}
			cui_clip = NULL;
			cui_rects_free(rs);
		}else{
			cui_draw(obj);
		}
	}
	if(obj == cui_focus_get() && !cui_is_visible_rect(obj)) cui_focus_set(NULL);
}

void
cui_x_set(cui obj, int x)
{
	cui_xywh_set(obj, x, obj->y, obj->w, obj->h);
}

void
cui_y_set(cui obj, int y)
{
	cui_xywh_set(obj, obj->x, y, obj->w, obj->h);
}

void
cui_xy_set(cui obj, int x, int y)
{
	cui_xywh_set(obj, x, y, obj->w, obj->h);
}

void
cui_w_set(cui obj, int w)
{
	cui_xywh_set(obj, obj->x, obj->y, w, obj->h);
}

void
cui_h_set(cui obj, int h)
{
	cui_xywh_set(obj, obj->x, obj->y, obj->w, h);
}

void
cui_wh_set(cui obj, int w, int h)
{
	cui_xywh_set(obj, obj->x, obj->y, w, h);
}

void
cui_wh_fit_chain(cui obj, cui child)
{
	if(child == NULL) return;
	if(obj->w < cui_x2(child)) obj->w = cui_x2(child);
	if(obj->h < cui_y2(child)) obj->h = cui_y2(child);
	cui_wh_fit_chain(obj, child->next);
}

void
cui_wh_fit(cui obj)
{
	int bak_w = obj->w, bak_h = obj->h;
	obj->w = obj->h = 0;
	cui_wh_fit_chain(obj, obj->children);
	if(obj->w != bak_w || obj->h != bak_h) cui_handler_call(obj, CUI_EVT_RESIZE, TRUE);
}

int
cui_index(cui obj)
{
	if(obj == NULL) return -1;
	return cui_index(obj->next) + 1;
}

cui
cui_index_to_child(cui obj, int i)
{
	cui c = obj->children;
	int ci = cui_index(c);

	if(ci < 0 || ci < i) return NULL;
	while(ci-- > i) c = c->next;
	return c;
}

void
cui_quit(void)
{
	cui_running_set(FALSE);
}

#define KEY_FD	0

static fd_set sel_rfs, res_rfs;
static int sel_inited = FALSE;

void
cui_select_init(void)
{
	FD_ZERO(&sel_rfs);
	FD_SET(KEY_FD, &sel_rfs);
	sel_inited = TRUE;
}

void
cui_select_add(int fd)
{
	FD_SET(fd, &sel_rfs);
}

void
cui_select_del(int fd)
{
	FD_CLR(fd, &sel_rfs);
}

int
cui_select(void)
{
	int nfds = 0, fd;
	struct timeval tm;

	res_rfs = sel_rfs;
	for(fd=0; fd<FD_SETSIZE; fd++) if(FD_ISSET(fd, &res_rfs) && fd > nfds) nfds = fd;
	tm.tv_sec = 0;
	tm.tv_usec = 100*1000;
	return select(nfds+1, &res_rfs, NULL, NULL, &tm);
}

int
cui_select_chk(int fd)
{
	return FD_ISSET(fd, &res_rfs);
}

typedef struct cui_readable{
	cui obj;
	int fd;
} *cui_readable;

static cui_list readable_list = NULL;

static void
cui_readable_select_setup(void)
{
	cui_list lp = readable_list;
	cui_select_init();
	for(; lp; lp=lp->next){
		cui_readable p = (cui_readable)lp->data;
		cui_select_add(p->fd);
	}
}

void
cui_readable_add(cui obj, int fd)
{
	cui_readable p = cui_list_add(&readable_list, sizeof(struct cui_readable));
	p->obj = obj;
	p->fd = fd;
	cui_readable_select_setup();
}

void
cui_readable_del(cui obj, int fd)
{
	cui_list lp = readable_list;
	cui_readable p;
	for(; lp; lp=lp->next){
		p = (cui_readable)lp->data;
		if(p->obj == obj && p->fd == fd) break;
	}
	if(lp == NULL) return;
	cui_list_del(&readable_list, p);
	cui_readable_select_setup();
}

void
cui_readable_add_bind(cui obj, int fd, cui_handler hdr, void *prm)
{
	cui_bind(obj, CUI_EVT_READABLE, hdr, prm);
	cui_readable_add(obj, fd);
}

static void
cui_readable_work(void)
{
	cui_list lp = readable_list, next;
	cui_readable p;
	for(; lp; lp=next){
		next = lp->next;
		p = (cui_readable)lp->data;
		if(!cui_select_chk(p->fd)) continue;
		cui_handler_call(p->obj, CUI_EVT_READABLE, p->fd);
	}
}

void
cui_main(cui top_obj, cui init_focus)
{
	struct cui_main_stack stk;
	cui_main_stack sp = (cui_main_stack)&stk;

	int key, val;
	cui focus, obj;

	sp->focus = init_focus;
	sp->running = TRUE;
	sp->prev = main_stack;
	main_stack = sp;

	cui_esc_enter();
	cui_key_enter();

	cui_clear(top_obj);
	cui_draw(top_obj);

	if(!sel_inited) cui_select_init();

	while(cui_running_get()){
		cui_timer_work();
		if(cui_select() <= 0) continue;
		cui_readable_work();
		if(!cui_select_chk(KEY_FD)) continue;
		if((key = cui_key_get2()) == 0) continue;
		focus = cui_focus_get();
		if(focus){
			if(cui_handler_call(focus, CUI_EVT_KEY, key)) continue;
		}
		obj = cui_focus_move_judge(top_obj, key, NULL, &val);
		if(obj){
			cui_focus_set(obj);
			cui_draw(focus);
			cui_draw(obj);
		}
	}
	cui_key_exit();
	cui_esc_exit();

	main_stack = sp->prev;
}

void
cui_del_from(cui obj, cui *from)
{
	if(*from == NULL) return;
	if(*from == obj){
		*from = obj->next;
		return;
	}
	cui_del_from(obj, &(*from)->next);
}

void
cui_del(cui obj)
{
	cui_del_from(obj, &obj->parent->children);
}

void
cui_free_chain(cui obj)
{
	if(obj == NULL) return;
	cui_free_chain(obj->next);
	cui_free(obj);
}

void
cui_free(cui obj)
{
	if(obj == NULL) return;
	cui_free_chain(obj->children);
	cui_list_del_all(&obj->handler_list);
	free(obj);
}

/* EOF */
